@@ -3,9 +3,9 @@
  *
  * Common functions for SAKURA and TOMOYO.
  *
- * Copyright (C) 2005-2011  NTT DATA CORPORATION
+ * Copyright (C) 2005-2012  NTT DATA CORPORATION
  *
- * Version: 1.6.9+   2011/05/05
+ * Version: 1.6.9+   2012/05/05
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -2566,7 +2566,11 @@ void ccs_load_policy(const char *filename)
 		envp[0] = "HOME=/";
 		envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
 		envp[2] = NULL;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 23) || defined(UMH_WAIT_PROC)
+		call_usermodehelper(argv[0], argv, envp, UMH_WAIT_PROC);
+#else
 		call_usermodehelper(argv[0], argv, envp, 1);
+#endif
 	}
 #elif defined(TASK_DEAD)
 	{
@@ -2608,10 +2612,10 @@ void ccs_load_policy(const char *filename)
 	}
 #endif
 #ifdef CONFIG_SAKURA
-	printk(KERN_INFO "SAKURA: 1.6.9+   2011/05/05\n");
+	printk(KERN_INFO "SAKURA: 1.6.9+   2012/05/05\n");
 #endif
 #ifdef CONFIG_TOMOYO
-	printk(KERN_INFO "TOMOYO: 1.6.9+   2011/05/05\n");
+	printk(KERN_INFO "TOMOYO: 1.6.9+   2012/05/05\n");
 #endif
 	printk(KERN_INFO "Mandatory Access Control activated.\n");
 	ccs_policy_loaded = true;
@@ -2763,36 +2767,17 @@ int ccs_check_supervisor(struct ccs_request_info *r, const char *fmt, ...)
  * ccs_poll_query - poll() for /proc/ccs/query.
  *
  * @file: Pointer to "struct file".
- * @wait: Pointer to "poll_table".
+ * @wait: Pointer to "poll_table". May be NULL.
  *
- * Returns POLLIN | POLLRDNORM when ready to read, 0 otherwise.
- *
- * Waits for access requests which violated policy in enforcing mode.
+ * Returns POLLIN | POLLRDNORM if ready to read a query, 0 otherwise.
  */
-static int ccs_poll_query(struct file *file, poll_table *wait)
+static unsigned int ccs_poll_query(struct file *file, poll_table *wait)
 {
-	struct list_head *tmp;
-	bool found = false;
-	u8 i;
-	for (i = 0; i < 2; i++) {
-		/***** CRITICAL SECTION START *****/
-		spin_lock(&ccs_query_list_lock);
-		list_for_each(tmp, &ccs_query_list) {
-			struct ccs_query_entry *ptr
-				= list_entry(tmp, struct ccs_query_entry, list);
-			if (ptr->answer)
-				continue;
-			found = true;
-			break;
-		}
-		spin_unlock(&ccs_query_list_lock);
-		/***** CRITICAL SECTION END *****/
-		if (found)
-			return POLLIN | POLLRDNORM;
-		if (i)
-			break;
-		poll_wait(file, &ccs_query_wait, wait);
-	}
+	if (!list_empty(&ccs_query_list))
+		return POLLIN | POLLRDNORM;
+	poll_wait(file, &ccs_query_wait, wait);
+	if (!list_empty(&ccs_query_list))
+		return POLLIN | POLLRDNORM;
 	return 0;
 }
 
@@ -2821,8 +2806,6 @@ static int ccs_read_query(struct ccs_io_buffer *head)
 	list_for_each(tmp, &ccs_query_list) {
 		struct ccs_query_entry *ptr
 			= list_entry(tmp, struct ccs_query_entry, list);
-		if (ptr->answer)
-			continue;
 		if (pos++ != head->read_step)
 			continue;
 		len = ptr->query_len;
@@ -2843,8 +2826,6 @@ static int ccs_read_query(struct ccs_io_buffer *head)
 	list_for_each(tmp, &ccs_query_list) {
 		struct ccs_query_entry *ptr
 			= list_entry(tmp, struct ccs_query_entry, list);
-		if (ptr->answer)
-			continue;
 		if (pos++ != head->read_step)
 			continue;
 		/*
@@ -2899,8 +2880,12 @@ static int ccs_write_answer(struct ccs_io_buffer *head)
 			= list_entry(tmp, struct ccs_query_entry, list);
 		if (ptr->serial != serial)
 			continue;
-		if (!ptr->answer)
-			ptr->answer = answer;
+		ptr->answer = answer;
+		/* Remove from ccs_query_list. */
+		if (ptr->answer) {
+			list_del(&ptr->list);
+			INIT_LIST_HEAD(&ptr->list);
+		}
 		break;
 	}
 	spin_unlock(&ccs_query_list_lock);
@@ -3234,12 +3219,12 @@ int ccs_open_control(const u8 type, struct file *file)
  * /proc/ccs/grant_log and /proc/ccs/reject_log are handled by
  * /usr/lib/ccs/ccs-auditd.
  */
-int ccs_poll_control(struct file *file, poll_table *wait)
+unsigned int ccs_poll_control(struct file *file, poll_table *wait)
 {
 	struct ccs_io_buffer *head = file->private_data;
 	if (!head->poll)
-		return -ENOSYS;
-	return head->poll(file, wait);
+		return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
+	return head->poll(file, wait) | POLLOUT | POLLWRNORM;
 }
 
 /**
